/*
 * Partially implemented disassembler for Samsung SSP160x processors
 * (C) Copyright 2007-2008, Grazvydas "notaz" Ignotas
 */

#include "ssp.hpp"

//----------------------------------------------------------------------
// Emulate an instruction
// This function should:
//      - create all xrefs from the instruction
//      - perform any additional analysis of the instruction/program
//        and convert the instruction operands, create comments, etc.
//      - create stack variables
//      - analyze the delayed branches and similar constructs
// The kernel calls ana() before calling emu(), so you may be sure that
// the 'cmd' structure contains a valid and up-to-date information.
// You are not allowed to modify the 'cmd' structure.
// Upon entering this function, the 'uFlag' variable contains the flags of
// cmd.ea. If you change the characteristics of the current instruction, you
// are required to refresh 'uFlag'.
// Usually the kernel calls emu() with consecutive addresses in cmd.ea but
// you can't rely on this - for example, if the user asks to analyze an
// instruction at arbirary address, his request will be handled immediately,
// thus breaking the normal sequence of emulation.
// If you need to analyze the surroundings of the current instruction, you
// are allowed to save the contents of the 'cmd' structure and call ana().
// For example, this is a very common pattern:
//  {
//    insn_t saved = cmd;
//    if ( decode_prev_insn(cmd.ea) != BADADDR )
//    {
//      ....
//    }
//    cmd = saved;
//  }

int idaapi emu(void)
{
  bool flow;               // does the current instruction pass
                           // execution to the next instruction?

  // setup
  ulong Feature = cmd.get_canon_feature();
  flow = true;

  // disable flow if CF_STOP set
  if (Feature & CF_STOP) {
    flow = false;
  }

  // check for call/branch
  if (cmd.Op2.type == o_near) {
    bool iscall = InstrIsSet(cmd.itype, CF_CALL);
    ua_add_cref(cmd.Op2.offb, cmd.Op2.addr, iscall ? fl_CN : fl_JN);
    // unconditional jump?
    if (!iscall && cmd.Op1.type == o_idpspec0 && cmd.Op1.specval == 0)
      flow = false;
  }

  // modifying PC?
  if (cmd.Op1.type == o_reg && cmd.Op1.reg == rPC) {
    flow = false;
  }

  // the immediate might actually be a address, so output byte address as a comment
  if (cmd.Op2.type == o_imm && cmd.Op2.value >= 0x400 && cmd.itype != SSP_cmpi) {
    char buff[16];
    qsnprintf(buff, sizeof(buff), "gloc_%04x", cmd.Op2.value << 1);
    set_cmt(cmd.ea, buff, 0);
    set_name(cmd.Op2.value << 1, buff);
  }

  // if the execution flow is not stopped here, then create
  // a xref to the next instruction.
  // Thus we plan to analyze the next instruction.
  if (flow) {
    ua_add_cref(0, cmd.ea + cmd.size, fl_F);
  }

  // OK (actual code unimportant)
  return 1;
}

