/*
 * Partially implemented disassembler for Samsung SSP160x processors
 * (C) Copyright 2007-2008, Grazvydas "notaz" Ignotas
 */

#include "ssp.hpp"
#include <fpro.h>
#include <diskio.hpp>


//--------------------------------------------------------------------------
// generate start of the disassembly
void idaapi header(void) {
  // generate standard header
  gen_cmt_line("Processor:        %s", inf.procName);
  gen_cmt_line("Target assembler: %s", ash.name);

  // output assembler-specific header
  if (ash.header != NULL) {
    for (const char **ptr=ash.header; *ptr != NULL; ptr++) {
      MakeLine(*ptr, 0);
    }
  }
}

// generate start of segment
void idaapi segstart(ea_t ea) {
  // generate ORG directive if necessary
  if (inf.s_org)
  {
    // get segment data
    segment_t *Sarea = getseg(ea);
    ulong org = ea - get_segm_base(Sarea);

    // generate line
    if ( org != 0 )
    {
      char buf[MAX_NUMBUF];
      btoa(buf, sizeof(buf), org);
      printf_line(inf.indent, COLSTR("%s %s", SCOLOR_ASMDIR), ash.origin, buf);
    }
  }
}

// generate end of the disassembly
void idaapi footer(void) {
  char buf[MAXSTR];

  // if assembler supplies end statement, output it
  if (ash.end != NULL) {
    MakeNull();
    tag_addstr(buf, buf+sizeof(buf), COLOR_ASMDIR, ash.end);
    MakeLine(buf, inf.indent);
  }
}

// --------------------------------------------------------------------------

void OutAddr(op_t& x) {
  // try and find the real name expression
  if ( !out_name_expr(x, x.addr, x.addr) )
  {
    // if name wasn't found, just output the value & add to noname queue
    OutValue(x, OOF_ADDR | OOFW_16 | OOFS_NOSIGN);
    QueueMark(Q_noName, cmd.ea);
  }
}

// --------------------------------------------------------------------------

static const char *ssp_cond_names[16] =
{
	"always",   "RESERVED", "gpi0",     "gpi1",
	"l",        "z",        "ov",       "n",
	"diof",     "gpi2",     "gpi3",     "RESERVED",
	"RESERVED", "RESERVED", "RESERVED", "RESERVED",
};

static const char *ssp_modifiers[4] =
{
	"", "+!", "-", "+"
};

static const char *ssp_modifiers2[4] =
{
	"|00", "|01", "|10", "|11"
};

static const char *ssp_acc_op[8] =
{
	"ror", "rol", "shr", "shl", "inc", "dec", "neg", "abs"
};

static const char *ssp_flag_op[16] =
{
	"?",     "?",     "resl", "setl", "resie", "setie", "?",   "?",
	"resop", "setop", "?",    "?",    "?",     "?",     "res", "set"
};

//----------------------------------------------------------------------
// generate the text representation of an operand

bool idaapi outop(op_t &x)
{
  char buf[64];

  // output operands
  switch (x.type)
  {
    case o_reg:
      if (x.reg > 48)
           qstrncpy(buf, "???", sizeof(buf));
      else qsnprintf(buf, sizeof(buf), "%s", ssp_reg_names[x.reg]);
      out_register(buf);
      break;

    case o_imm:
      OutValue(x);
      break;

    case o_near:
      OutAddr(x);
      break;

    case o_idpspec0: // guessed conds
      out_snprintf("%s", ssp_cond_names[x.specval&0xf]);
      if (x.specval > 0)
        out_snprintf("=%i", x.specflag1);
      break;

    case o_idpspec1: { // ri|mod
      const char *pr = "", *po = "";
      const char **mods = (x.specval == 3 || x.specval == 7) ? ssp_modifiers2 : ssp_modifiers;
      if (x.specflag2 == 1) { pr =  "("; po =  ")"; }
      if (x.specflag2 == 2) { pr = "(("; po = "))"; }
      out_snprintf("%s%s%s%s", pr, ssp_reg_names[(x.specval&7)|0x20], mods[x.specflag1&3], po);
      break;
    }

    case o_idpspec2: // (a)
      out_snprintf("(A)");
      break;

    case o_idpspec3: // mod op (A operations)
      out_snprintf("%s", ssp_acc_op[x.specval&7]);
      break;

    case o_idpspec4: // mpys/mpya/mld 'b' operand
      out_snprintf("%i", x.specval&1);
      break;

    case o_idpspec5: // "mod f" operands
      if (x.specval >= 16)
        out_snprintf("f");
      else
        out_snprintf("%s", ssp_flag_op[x.specval]);
      break;

    default:
      qstrncpy(buf, "???", sizeof(buf));
      out_line(buf, COLOR_DEFAULT);
      break;
  }

  // OK
  return 1;
}


//----------------------------------------------------------------------
// generate a text representation of an instruction
// the information about the instruction is in the 'cmd' structure
void idaapi out(void)
{
  char buf[MAXSTR];

  // setup the output pointer
  init_output_buffer(buf, sizeof(buf));

  // output instruction mnemonics
  OutMnem();

  if (cmd.itype == SSP_unknown)
  {
    OutBadInstruction();
    //OutLong(cmd.auxpref, 16);
    out_snprintf("%04x", cmd.auxpref);
    goto finish;
  }

  //  out_symbol(',');
  //  OutChar(' ');

  // output the first operand
  if (cmd.Op1.type != o_void) {
    out_one_operand(0);
  }

  // output the second operand
  if (cmd.Op2.type != o_void) {
    out_symbol(',');
    OutChar(' ');
    out_one_operand(1);
  }

  if (cmd.Op3.type != o_void) {
    out_symbol(',');
    OutChar(' ');
    out_one_operand(2);
  }


finish:
  // terminate the output string
  term_output_buffer();

  // ask to attach a possible user-defined comment to it
  gl_comm = 1;

  // pass the generated line to the kernel
  MakeLine(buf);
}



// --------------------------------------------------------------------------
// customised address output
void idaapi out_data(ea_t ea)
{
  // if addres is valid, use normal output function
  if (isLoaded(ea)) {
    intel_data(ea);
  } else {
    MakeLine(COLSTR("; db ?", SCOLOR_SYMBOL));
  }
}
